#include <SPI.h>
#include <SD.h>
#include <Wire.h>
#include <MPU6050_tockn.h>
#include <Adafruit_BMP280.h>
#include <Servo.h>

#define button 7
#define buzzer 3
//Inside:
#define c1 4
//Outside:
#define c2 2
#define BMP_SCK 13
#define BMP_MISO 12
#define BMP_MOSI 11
#define BMP_CS 10

const float Xtune = 0, Ytune = 0;
Adafruit_BMP280 bmp;
MPU6050 mpu6050(Wire);

File data;

Servo servoX;
Servo servoY;

const double ServoXMult = 3, ServoYMult = 3, P = .3, D = .2;

float initial_altitude, current_altitude, highest_altitude, gyro_x, gyro_y, gyro_z, accel_x, accel_y, accel_z, angular_velocity_x, angular_velocity_y, temp, servo_x, servo_y, servo_tiltX, servo_tiltY;
bool log_data = true, apogee = false, failable = false;

void setup() {
  pinMode(button, INPUT_PULLUP);
  pinMode(buzzer, OUTPUT);
  pinMode(c1, OUTPUT);
  pinMode(c2, OUTPUT);
  servoX.attach(5);
  servoY.attach(6);
  Serial.begin(9600);
  digitalWrite(c1, LOW);
  digitalWrite(c2, LOW);

  bmp.begin();
  bmp.setSampling(Adafruit_BMP280::MODE_NORMAL, Adafruit_BMP280::SAMPLING_X2, Adafruit_BMP280::SAMPLING_X16, Adafruit_BMP280::FILTER_X16, 1);

  Wire.begin();
  mpu6050.begin();


  while (!SD.begin(10)) {
    tone(buzzer, 600);
    delay(200);
    noTone(buzzer);
    delay(200);
    tone(buzzer, 450);
    delay(200);
    noTone(buzzer);
  }

  //calibrate();
}

void calibrate() {
  while (digitalRead(button) == LOW) {}
  servoX.write(90 + Xtune);
  servoY.write(90 + Ytune);
  tone(buzzer, 261);
  delay(200);
  noTone(buzzer);
  tone(buzzer, 329);
  delay(200);
  noTone(buzzer);
  tone(buzzer, 392);
  delay(200);
  noTone(buzzer);
  delay(2000);

  mpu6050.calcGyroOffsets(true);
  initial_altitude = bmp.readAltitude(1013.25);

  tone(buzzer, 659);
  delay(200);
  noTone(buzzer);
  tone(buzzer, 784);
  delay(200);
  noTone(buzzer);
  tone(buzzer, 1047);
  delay(200);
  noTone(buzzer);
}

void beep(int frequency, int duration) {
  tone(buzzer, frequency);
  delay(duration);
  noTone(buzzer);
}

bool handleButtonPresses() {
  bool buttonPressed = false;
  unsigned long lastPressTime = 0;
  unsigned long debounceDelay = 20;
  unsigned long cooldownDuration = 60;
  unsigned int buttonCounter = 0;
  unsigned long startTime = millis();

  while (millis() - startTime < 2000) {
    int buttonState = digitalRead(button);

    if (buttonState == HIGH && millis() - lastPressTime >= debounceDelay && !buttonPressed) {
      buttonPressed = true;
      beep(500, 100);
      lastPressTime = millis();
      buttonCounter++;
      delay(cooldownDuration);
    }

    if (buttonState == LOW && buttonPressed) {
      buttonPressed = false;
    }

    if (buttonCounter == 5 && (millis() - lastPressTime) < 2500) {
      buttonCounter = 0;
      return true;
    }
  }

  buttonCounter = 0;
  return false;
}

bool countdown() {
  unsigned long startTime = millis();
  unsigned long beepInterval = 900;
  unsigned long countdownDuration = 31800;
  unsigned long lastBeepTime = millis();
  fadeBuzzer(200, 2500, 2000, true);
  failable = true;
  while (millis() - startTime < countdownDuration) {
    if ((digitalRead(button) == HIGH || apogee == true)) {
      fadeBuzzer(200, 2500, 2000, false);
      apogee = false;
      return false;
    }
    sensors();
    if (millis() - lastBeepTime >= beepInterval) {
      if (millis() - startTime >= countdownDuration - 6000) {
        tone(buzzer, 2500);
        if ((digitalRead(button) == HIGH || apogee == true)) {
          fadeBuzzer(200, 2500, 2000, false);
          apogee = false;
          return false;
        }
      } else {
        beep(2500, 100);
      }
      lastBeepTime = millis();
    }
  }
  noTone(buzzer);
  data = SD.open("data.txt", FILE_WRITE);
  data.println("Launch!");
  data.close();
  digitalWrite(c1, HIGH);
  return true;
}


void sensors() {
  static unsigned long prevTime = 0; // Variable to store the previous time
  static float prev_gyro_x = 0;      // Variable to store the previous gyro_x value
  static float prev_gyro_y = 0;      // Variable to store the previous gyro_y value

  mpu6050.update();
  current_altitude = bmp.readAltitude(1013.25) - initial_altitude;
  temp = mpu6050.getTemp();
  gyro_x = mpu6050.getAngleX();
  gyro_y = mpu6050.getAngleY();
  gyro_z = mpu6050.getAngleZ();
  accel_x = mpu6050.getAccX();
  accel_y = mpu6050.getAccY();
  accel_z = mpu6050.getAccZ();
  servo_x = servo_tiltX;
  servo_y = servo_tiltY;

  // Calculate time elapsed since the last call to sensors()
  unsigned long currentTime = millis();
  unsigned long elapsedTime = currentTime - prevTime;

  // Check if at least 10 milliseconds have passed since the last calculation
  if (elapsedTime >= 10) {
    // Calculate angular velocity for gyro_x and gyro_y
    angular_velocity_x = (gyro_x - prev_gyro_x) / (elapsedTime / 1000.0); // Convert milliseconds to seconds
    angular_velocity_y = (gyro_y - prev_gyro_y) / (elapsedTime / 1000.0); // Convert milliseconds to seconds

    // Update previous gyro values and time for the next iteration
    prev_gyro_x = gyro_x;
    prev_gyro_y = gyro_y;
    prevTime = currentTime;
  }

  if (log_data) {
    data = SD.open("data.txt", FILE_WRITE);
    data.print(current_altitude);
    data.print(",");
    data.print(temp);
    data.print(",");
    data.print(gyro_x);
    data.print(",");
    data.print(gyro_y);
    data.print(",");
    data.print(gyro_z);
    data.print(",");
    data.print(accel_x);  
    data.print(",");
    data.print(accel_y);
    data.print(",");
    data.print(accel_z);
    data.print(",");
    data.print(servo_x);
    data.print(",");
    data.print(servo_y);
    data.println();
    data.close();
    if ((gyro_x > 70 || gyro_x < -70 || gyro_y > 70 || gyro_y < -70) && failable) {
      apogee = true;
    }
    if (current_altitude < highest_altitude - 1) {
      apogee = true;
    }
    if (apogee) {
      return;
    }
    if (current_altitude > highest_altitude) {
      highest_altitude = current_altitude;
    }
  }
  log_data = !log_data;
}



void fadeBuzzer(int startFrequency, int endFrequency, int totalDuration, bool fadeUp) {

  int steps = abs(endFrequency - startFrequency) / 10;
  int stepDuration = totalDuration / steps;

  if (fadeUp) {
    for (int frequency = startFrequency; frequency <= endFrequency; frequency += 10) {
      beep(frequency, stepDuration);
      sensors();
    }
  } else {
    for (int frequency = endFrequency; frequency >= startFrequency; frequency -= 10) {
      beep(frequency, stepDuration);
    }
  }
}

void land() {
  digitalWrite(c1, LOW);
  digitalWrite(c2, HIGH);
  servoX.write(90 + Xtune);
  servoY.write(90 + Ytune);
  data = SD.open("data.txt", FILE_WRITE);
  data.println("Parachute deployed");
  data.close();
  unsigned long lastBeepTime = millis();
  while (true) {
    // Beep for 500ms every second
    if (millis() - lastBeepTime >= 1000) {
      tone(buzzer, 2500);
      delay(500);
      noTone(buzzer);
      lastBeepTime = millis();
    }
    sensors();
  }
}

float mapFloat(float x, float in_min, float in_max, float out_min, float out_max) {
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

void TVC(){
  sensors();
  float motor_tiltX = (P * gyro_x) + (D * angular_velocity_x);

  servo_tiltX = constrain(motor_tiltX,-5,5) * ServoXMult;

  servo_tiltX = mapFloat(servo_tiltX, -90.0, 90.0, 0.0, 180.0) + Xtune;

  servoX.write(servo_tiltX);

  float motor_tiltY = (P * gyro_y) + (D * angular_velocity_y);s

  servo_tiltY = constrain(motor_tiltY,-5,5) * ServoYMult;

  servo_tiltY = mapFloat(servo_tiltY, -90.0, 90.0, 0.0, 180.0) + Ytune;

  servoY.write(servo_tiltY);
}


void loop() {
  while (handleButtonPresses() == false) {}
  if (countdown()) {
    while (apogee == false) {
      TVC();
    }
    land();
  }
}

// servoX on inside, servoY on outside of PCB